// Java Authentication System
// JarFile Class

import java.util.zip.*;
import java.util.*;
import java.io.*;
import java.security.*;
import sun.misc.BASE64Decoder;

public class JarFile extends ZipFile {
	// A signed JarFile should have the following entries
	// - One MANIFEST.MF file
	// - One or more signature instruction files (xxx.SF)
	// - One or more digital signature files (xxx.suffix) Where suffix is 
	//	 the name of the digital signature algorithm
	ManifestFile manifest = null;
	SignatureFile sigfile = null;
	DigitalSignatureFile digsigfile = null;
	FileEntry fileentry = null;

	Vector sigfiles = new Vector(1);
	Vector digsigfiles = new Vector(1);
	Vector fileentries = new Vector(50);

	int numOfFiles=0; // Number of files in archive (Files in META-INF are not counted)

	JarFile(File file) throws IOException, ZipException {
		super( file );
		build();
	}

	JarFile(String string) throws IOException {
		super( string );
		build();
	}

	public boolean checkIntegrity()
	{
		if ( manifest == null ) {
			JasDebug.println("Error:	No manifest file found!");
			return false;
		}

		if ( checkFileIntegrity() == false ) {
			JasDebug.println("Error:	File integrity check failed!");
			return false;
		}
		
		// Check the manifest file for each signer
		for (Enumeration sfs = sigfiles.elements() ; sfs.hasMoreElements() ;) {
           		SignatureFile current = (SignatureFile)	sfs.nextElement();
				if ( current.checkManifestIntegrity( manifest ) == false ) {
					JasDebug.println("Error:	Manifest integrity check failed!");
					return false;
				}
		}

		JasDebug.println("Succeed! Archive passed integrity check.");
		return true;
	}

	private boolean checkFileIntegrity() {

		// For each entry in the manifest file:
		//		1. Find the corresponding file in the archive.
		//
		//		2. Compute a digest for the file and compare
		//		to the digest stored in the manifest file
	
		ContentSection mfentry = null;
		Vector v = manifest.getContent();
		BASE64Decoder decoder = new BASE64Decoder();
		int passed = 0;		// Number of files that passed the integrity check

		JasDebug.println("Starting file integrity check");
	
		for (Enumeration mfes = v.elements() ; mfes.hasMoreElements() ;) {
           	mfentry = (ContentSection) mfes.nextElement();
			String name = mfentry.getName();
			DigestData[] digests = mfentry.getDigests();
			
			// Get the corresponding file

			JasDebug.println("Now checking: " + name);
			for (Enumeration fes = fileentries.elements() ; fes.hasMoreElements() ;) {
           		fileentry = (FileEntry)	fes.nextElement();
				
				if( name.equals(fileentry.getName()) ) {
					JasDebug.println("Found a file that matches");
					try {	
						// Check if any of the stored hashes is of a supported type
						for( int i=0; i < digests.length; i++) {
							if ( (JasConfig.supportedHashAlgorithm( digests[i].algorithm )) == true ) {
								JasDebug.println(digests[i].algorithm + " is a supported algorithm");
								byte[] filedigest = fileentry.getDigest( digests[i].algorithm );
								String digestString = new String( digests[i].digest );
								byte[] manifestdigest = decoder.decodeBuffer( digestString );
								if ( (MessageDigest.isEqual( filedigest, manifestdigest )) == true ) {
									JasDebug.println("Digests match!");
									passed++;
									break;
								}							
							}
						}
					} catch ( NoSuchAlgorithmException e ) {
						JasDebug.println("NoSuchAlgorithmException");
						return false;
					} catch ( IOException e ) {
						JasDebug.println("NoSuchAlgorithmException");
						return false;
					}
					break;
				}
			}
		}
	
		if ( passed == v.size() )
			return true;
		else
			return false;
	}

	private void build() throws IOException {
		InputStream in;

		for( Enumeration jarEntries = this.entries(); jarEntries.hasMoreElements(); ) {
			ZipEntry current = (ZipEntry) jarEntries.nextElement();
			int size = (int)current.getSize();
			String name = current.getName();			
		
			if ( name.startsWith("META-INF/") ) {
		
				if ( name.equalsIgnoreCase("META-INF/MANIFEST.MF") ) {
					in = this.getInputStream( current );
					manifest = new ManifestFile( in, size );
				} else

				if ( (name.endsWith(".SF")) || (name.endsWith(".sf")) ) {
					in = this.getInputStream( current );
					sigfile = new SignatureFile( in, size );
					sigfiles.addElement( sigfile );
				} else

				if ( (name.endsWith(".DSA")) || (name.endsWith(".RSA")) ||
					(name.endsWith(".dsa")) || (name.endsWith(".rsa")) ) {
					in = this.getInputStream( current );
					digsigfile = new DigitalSignatureFile( in );
					digsigfiles.addElement( digsigfile );
				} else {
					// throw Exception("JAS:	Authentication failed. There are untrusted files in META-INF");					
					System.out.println("JAS Failed:	Unknown file type in META-INF");
					throw new IOException();	// XXX XXX throw something else!!!
				}			
			} else {
				numOfFiles++;
				in = this.getInputStream( current );
				fileentry = new FileEntry( name, in, size );
				fileentries.addElement( fileentry );
			}
		} // FOR Loop


// XXX do this elsewhere, not in this function
/*
		if ( manifest == null ) {
			System.out.println("JAS Error:	No manifest file found!");
			// XXX Remove
			System.exit(1);
		}

		for (Enumeration sfs = sigfiles.elements() ; sfs.hasMoreElements() ;) {
            SignatureFile sfile = (SignatureFile) sfs.nextElement();
			if ( sfile == null ) {
				System.out.println("JAS Error:	No signature file found!");
				// XXX Remove
				System.exit(1);
			}
			System.out.println("JAS Debug:	Signature file found!");
			
			// Compare digest stored for each manifest file entry in
			// signature file, to a digest calculated now for that
			// particular entry.
			Vector v = sfile.getContent();
			for (Enumeration e = v.elements() ; e.hasMoreElements() ;) {
            	ContentSection section = (ContentSection) e.nextElement();			
				DigestData dig[] = section.getDigests();
				JasDebug.println("Name: " + section.getName() );
				for ( int k=0 ; k < dig.length ; k++ ) {
					JasDebug.println("Algorithm:" + dig[k].algorithm);
				}
			}
			if (sfile.compareDigests( manifest ) == false ) {
				System.out.println("JAS Error:	Digests differs (SF -> MF)");
				// XXX Handle this situation
			}
		}
*/
	}
}
